/*
 * File: z_en_horse_ganon.c
 * Overlay: ovl_En_Horse_Ganon
 * Description: Ganondorf's Horse
 */

#include "z_en_horse_ganon.h"

#include "libc64/math64.h"
#include "gfx.h"
#include "gfx_setupdl.h"
#include "ichain.h"
#include "sfx.h"
#include "sys_math3d.h"
#include "z_lib.h"
#include "z64play.h"
#include "z64player.h"
#include "z64skin.h"

#include "assets/objects/object_horse_ganon/object_horse_ganon.h"

#define FLAGS ACTOR_FLAG_UPDATE_CULLING_DISABLED

typedef struct unk_D_80A69248 {
    /* 0x0 */ Vec3s unk_0;
    /* 0x6 */ u8 unk_6;
} unk_D_80A69248; // size = 0x8

void EnHorseGanon_Init(Actor* thisx, PlayState* play);
void EnHorseGanon_Destroy(Actor* thisx, PlayState* play);
void EnHorseGanon_Update(Actor* thisx, PlayState* play);
void EnHorseGanon_Draw(Actor* thisx, PlayState* play);

void func_80A68AC4(EnHorseGanon* this);
void func_80A68AF0(EnHorseGanon* this, PlayState* play);
void func_80A68DB0(EnHorseGanon* this, PlayState* play);

ActorProfile En_Horse_Ganon_Profile = {
    /**/ ACTOR_EN_HORSE_GANON,
    /**/ ACTORCAT_BG,
    /**/ FLAGS,
    /**/ OBJECT_HORSE_GANON,
    /**/ sizeof(EnHorseGanon),
    /**/ EnHorseGanon_Init,
    /**/ EnHorseGanon_Destroy,
    /**/ EnHorseGanon_Update,
    /**/ EnHorseGanon_Draw,
};

static AnimationHeader* sAnimations[] = {
    &gHorseGanonIdleAnim,     &gHorseGanonWhinnyAnim,    &gHorseGanonWalkingAnim,
    &gHorseGanonTrottingAnim, &gHorseGanonGallopingAnim, &gHorseGanonRearingAnim,
};

static f32 splaySpeeds[] = { 2.0f / 3.0f, 2.0f / 3.0f, 1.0f, 1.0f, 1.0f, 2.0f / 3.0f };

static ColliderCylinderInit sCylinderInit = {
    {
        COL_MATERIAL_NONE,
        AT_NONE,
        AC_NONE,
        OC1_ON | OC1_TYPE_ALL,
        OC2_TYPE_1 | OC2_UNK1,
        COLSHAPE_CYLINDER,
    },
    {
        ELEM_MATERIAL_UNK0,
        { 0x00000000, 0x00, 0x00 },
        { 0x00000000, 0x00, 0x00 },
        ATELEM_NONE,
        ACELEM_NONE,
        OCELEM_ON,
    },
    { 40, 100, 0, { 0, 0, 0 } },
};

static ColliderJntSphElementInit sJntSphElementsInit[] = {
    {
        {
            ELEM_MATERIAL_UNK0,
            { 0x00000000, 0x00, 0x00 },
            { 0x00000000, 0x00, 0x00 },
            ATELEM_NONE,
            ACELEM_NONE,
            OCELEM_ON,
        },
        { 13, { { 0, 0, 0 }, 20 }, 100 },
    },
};

static ColliderJntSphInit sJntSphInit = {
    {
        COL_MATERIAL_NONE,
        AT_NONE,
        AC_ON | AC_TYPE_PLAYER,
        OC1_ON | OC1_TYPE_ALL,
        OC2_TYPE_1 | OC2_UNK1,
        COLSHAPE_JNTSPH,
    },
    1,
    sJntSphElementsInit,
};

static CollisionCheckInfoInit sColChkInfoInit = { 10, 35, 100, MASS_HEAVY };

static unk_D_80A69248 D_80A69248[] = {
    { 0x09B8, 0x0126, 0x0E2C, 0x07 }, { 0x0C11, 0x017A, 0x1269, 0x07 }, { 0x064E, 0xFEFB, 0x1DAC, 0x07 },
    { 0x02F2, 0xFF45, 0x244F, 0x07 }, { 0xF96E, 0xFE0C, 0x3122, 0x07 }, { 0xF328, 0xFE0C, 0x32D5, 0x07 },
    { 0xEBEA, 0xFE5F, 0x2D6E, 0x07 }, { 0xE95E, 0xFE27, 0x2565, 0x07 }, { 0xE593, 0xFE0C, 0x20AC, 0x07 },
    { 0xE625, 0xFE77, 0x1B07, 0x07 }, { 0xEBB7, 0x007C, 0x1539, 0x07 }, { 0xF466, 0x0002, 0x11B9, 0x07 },
    { 0xF47B, 0xFFDD, 0x11AF, 0x07 }, { 0xF88D, 0xFFD1, 0x0BA2, 0x07 },
};

static s32 D_80A692B8[] = { 0, 16 };

static InitChainEntry sInitChain[] = {
    ICHAIN_F32(cullingVolumeScale, 1200, ICHAIN_STOP),
};

static EnHorseGanonActionFunc sActionFuncs[] = { func_80A68AF0, func_80A68DB0 };

void func_80A68660(unk_D_80A69248* data, s32 index, Vec3f* vec) {
    vec->x = data[index].unk_0.x;
    vec->y = data[index].unk_0.y;
    vec->z = data[index].unk_0.z;
}

void func_80A686A8(EnHorseGanon* this, PlayState* play) {
    Vec3f* tempPos;
    Vec3f vec;
    s16 y;

    func_80A68660(D_80A69248, this->unk_1EC, &vec);
    if (Math3D_Vec3f_DistXYZ(&vec, &this->actor.world.pos) <= 400.0f) {
        this->unk_1EC++;
        if (this->unk_1EC >= 14) {
            this->unk_1EC = 0;
            func_80A68660(D_80A69248, 0, &vec);
        }
    }

    tempPos = &this->actor.world.pos;
    y = Math_Vec3f_Yaw(tempPos, &vec) - this->actor.world.rot.y;
    if (y >= 301) {
        this->actor.world.rot.y += 300;
    } else if (y < -300) {
        this->actor.world.rot.y -= 300;
    } else {
        this->actor.world.rot.y += y;
    }
    this->actor.shape.rot.y = this->actor.world.rot.y;

    if (Actor_WorldDistXZToActor(&this->actor, &GET_PLAYER(play)->actor) <= 300.0f) {
        if (this->actor.speed < 12.0f) {
            this->actor.speed += 1.0f;
        } else {
            this->actor.speed -= 1.0f;
        }
    } else if (this->actor.speed < D_80A69248[this->unk_1EC].unk_6) {
        this->actor.speed += 0.5f;
    } else {
        this->actor.speed -= 0.5f;
    }
}

void func_80A68870(EnHorseGanon* this) {
    if ((this->skin.skelAnime.curFrame > D_80A692B8[this->soundCount]) &&
        (this->soundCount != 0 || !(this->skin.skelAnime.curFrame > D_80A692B8[1]))) {
        Audio_PlaySfxGeneral(NA_SE_EV_HORSE_WALK, &this->actor.projectedPos, 4, &gSfxDefaultFreqAndVolScale,
                             &gSfxDefaultFreqAndVolScale, &gSfxDefaultReverb);

        this->soundCount++;
        if (this->soundCount >= 2) {
            this->soundCount = 0;
        }
    }
}

void EnHorseGanon_Init(Actor* thisx, PlayState* play) {
    EnHorseGanon* this = (EnHorseGanon*)thisx;

    Actor_ProcessInitChain(&this->actor, sInitChain);
    Actor_SetScale(&this->actor, 0.0115f);

    this->actor.gravity = -3.5f;

    ActorShape_Init(&this->actor.shape, 0.0f, ActorShadow_DrawHorse, 20.0f);
    this->actor.speed = 0.0f;
    this->actor.focus.pos = this->actor.world.pos;
    this->action = 0;
    this->actor.focus.pos.y += 70.0f;
    Skin_Init(play, &this->skin, &gHorseGanonSkel, &gHorseGanonIdleAnim);
    this->currentAnimation = 0;
    Animation_PlayOnce(&this->skin.skelAnime, sAnimations[0]);

    Collider_InitCylinder(play, &this->bodyCollider);
    Collider_SetCylinder(play, &this->bodyCollider, &this->actor, &sCylinderInit);
    Collider_InitJntSph(play, &this->headCollider);
    Collider_SetJntSph(play, &this->headCollider, &this->actor, &sJntSphInit, this->headColliderElements);

    CollisionCheck_SetInfo(&this->actor.colChkInfo, NULL, &sColChkInfoInit);
    func_80A68AC4(this);
}

void EnHorseGanon_Destroy(Actor* thisx, PlayState* play) {
    EnHorseGanon* this = (EnHorseGanon*)thisx;

    Skin_Free(play, &this->skin);
    Collider_DestroyCylinder(play, &this->bodyCollider);
    Collider_DestroyJntSph(play, &this->headCollider);
}

void func_80A68AC4(EnHorseGanon* this) {
    this->action = 0;
    Animation_PlayLoop(&this->skin.skelAnime, sAnimations[4]);
}

void func_80A68AF0(EnHorseGanon* this, PlayState* play) {
    this->actor.speed = 0.0f;
    SkelAnime_Update(&this->skin.skelAnime);
}

void func_80A68B20(EnHorseGanon* this) {
    s32 animationChanged;
    f32 sp30;

    animationChanged = 0;
    this->action = 1;
    if (this->actor.speed <= 3.0f) {
        if (this->currentAnimation != 2) {
            animationChanged = 1;
        }
        this->currentAnimation = 2;
    } else if (this->actor.speed <= 6.0f) {
        if (this->currentAnimation != 3) {
            animationChanged = 1;
        }
        this->currentAnimation = 3;
    } else {
        if (this->currentAnimation != 4) {
            animationChanged = 1;
        }
        this->currentAnimation = 4;
    }

    if (this->currentAnimation == 2) {
        sp30 = this->actor.speed / 3.0f;
    } else if (this->currentAnimation == 3) {
        sp30 = this->actor.speed / 5.0f;
        Audio_PlaySfxGeneral(NA_SE_EV_HORSE_RUN, &this->actor.projectedPos, 4, &gSfxDefaultFreqAndVolScale,
                             &gSfxDefaultFreqAndVolScale, &gSfxDefaultReverb);
    } else if (this->currentAnimation == 4) {
        sp30 = this->actor.speed / 7.0f;
        Audio_PlaySfxGeneral(NA_SE_EV_HORSE_RUN, &this->actor.projectedPos, 4, &gSfxDefaultFreqAndVolScale,
                             &gSfxDefaultFreqAndVolScale, &gSfxDefaultReverb);
    } else {
        sp30 = 1.0f;
    }

    if (animationChanged == 1) {
        Animation_Change(&this->skin.skelAnime, sAnimations[this->currentAnimation],
                         splaySpeeds[this->currentAnimation] * sp30 * 1.5f, 0.0f,
                         Animation_GetLastFrame(sAnimations[this->currentAnimation]), ANIMMODE_ONCE, -3.0f);
    } else {
        Animation_Change(&this->skin.skelAnime, sAnimations[this->currentAnimation],
                         splaySpeeds[this->currentAnimation] * sp30 * 1.5f, 0.0f,
                         Animation_GetLastFrame(sAnimations[this->currentAnimation]), ANIMMODE_ONCE, 0.0f);
    }
}

void func_80A68DB0(EnHorseGanon* this, PlayState* play) {
    if (this->currentAnimation == 2) {
        func_80A68870(this);
    }

    func_80A686A8(this, play);

    if (SkelAnime_Update(&this->skin.skelAnime)) {
        func_80A68B20(this);
    }
}

void func_80A68E14(EnHorseGanon* this, PlayState* play) {
    s32 pad;
    CollisionPoly* poly;
    s32 pad2;
    Vec3f checkPos;
    s32 bgId;

    checkPos.x = Math_SinS(this->actor.shape.rot.y) * 30.0f + this->actor.world.pos.x;
    checkPos.y = this->actor.world.pos.y + 60.0f;
    checkPos.z = Math_CosS(this->actor.shape.rot.y) * 30.0f + this->actor.world.pos.z;
    this->unk_1F4 = BgCheck_EntityRaycastDown3(&play->colCtx, &poly, &bgId, &checkPos);
    this->actor.shape.rot.x = RAD_TO_BINANG(Math_FAtan2F(this->actor.world.pos.y - this->unk_1F4, 30.0f));
}

void EnHorseGanon_Update(Actor* thisx, PlayState* play) {
    EnHorseGanon* this = (EnHorseGanon*)thisx;
    s32 pad;

    sActionFuncs[this->action](this, play);
    Actor_MoveXZGravity(&this->actor);
    Actor_UpdateBgCheckInfo(play, &this->actor, 20.0f, 55.0f, 100.0f,
                            UPDBGCHECKINFO_FLAG_0 | UPDBGCHECKINFO_FLAG_2 | UPDBGCHECKINFO_FLAG_3 |
                                UPDBGCHECKINFO_FLAG_4);
    this->actor.focus.pos = this->actor.world.pos;
    this->actor.focus.pos.y += 70.0f;
    Collider_UpdateCylinder(&this->actor, &this->bodyCollider);
    CollisionCheck_SetOC(play, &play->colChkCtx, &this->bodyCollider.base);
}

void EnHorseGanon_PostDraw(Actor* thisx, PlayState* play, Skin* skin) {
    Vec3f sp4C;
    Vec3f sp40;
    EnHorseGanon* this = (EnHorseGanon*)thisx;
    s32 index;

    for (index = 0; index < this->headCollider.count; index++) {
        sp4C.x = this->headCollider.elements[index].dim.modelSphere.center.x;
        sp4C.y = this->headCollider.elements[index].dim.modelSphere.center.y;
        sp4C.z = this->headCollider.elements[index].dim.modelSphere.center.z;

        Skin_GetLimbPos(skin, this->headCollider.elements[index].dim.limb, &sp4C, &sp40);

        this->headCollider.elements[index].dim.worldSphere.center.x = sp40.x;
        this->headCollider.elements[index].dim.worldSphere.center.y = sp40.y;
        this->headCollider.elements[index].dim.worldSphere.center.z = sp40.z;

        this->headCollider.elements[index].dim.worldSphere.radius =
            this->headCollider.elements[index].dim.modelSphere.radius * this->headCollider.elements[index].dim.scale;
    }

    //! @bug see relevant comment in `EnHorse_SkinCallback1`
    CollisionCheck_SetOC(play, &play->colChkCtx, &this->headCollider.base);
}

void EnHorseGanon_Draw(Actor* thisx, PlayState* play) {
    EnHorseGanon* this = (EnHorseGanon*)thisx;

    func_80A68E14(this, play);
    Gfx_SetupDL_25Opa(play->state.gfxCtx);
    func_800A6330(&this->actor, play, &this->skin, EnHorseGanon_PostDraw, true);
}
